.586
.model flat, stdcall

option casemap: none

include macros.def
include ntstatus.inc
include ntddk.inc
include ntoskrnl.inc
include w2kundoc.inc

include strings.mac
include user32.inc
includelib user32.lib
includelib ntoskrnl.lib


comment	*
	;---------------------------------------------------------------------
		Coded by The Dude of Dudes.
		Thanks to Four-F for his Tutorials on kernel mode drivers
		and his help with Memory Management.
	;---------------------------------------------------------------------

	MemPhys DriverEntry code. Maps the Virtual address of a process specific page 
	to any physical page by modifying Page Tables.
	Maintains read/write access rights of Virtual page.
	

Use DeviceIOControl to access drivers functions:


IOCTL_GETPHYS
~~~~~~~~~~~~~

	Retrieves physical address of a process specific address.
	IN  - address to translate (dword)
	OUT - physical address (dword)

	
IOCTL_CHANGEPTE
~~~~~~~~~~~~~~~

	Maps a view of any physical page to a pre-committed process specific page.
	IN  - virtual address to map (dword), Physical address to view (dword)
	OUT - copy of virtual address (dword), original physical address (dword)

	Using Memory Management API's on the committed page while the Page Table
	entry is modified may cause system crash. Map commited page back to 
	default value before calling any Heap/Virtual functions or closing
	device handle.

	*

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;	   Prototypes
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	DriverEntry	proto, :dword, :dword
	DriverUnload	proto, :dword
	DispatchCreate	proto, :dword, :dword
	DispatchClose	proto, :dword, :dword
	DispatchControl proto, :dword, :dword
	ChangePTE	proto, :dword, :dword

.data

;==== Object Manager Name for our Device ====
	CCOUNTED_UNICODE_STRING		"\\Device\\PhysicalMemoryViewer", usDevice, 4

;==== Win32 API User mode name for Device ====	
	CCOUNTED_UNICODE_STRING		"\\??\\MemPhys", usSymbollic, 4


;==== User DeviceIO Control Values ====

	GETPHYS = 800h
	CHANGEPTE = 801h


;==== Control Code Definitions =====

	IOCTL_GETPHYS	equ CTL_CODE(FILE_DEVICE_UNKNOWN, GETPHYS, METHOD_BUFFERED, \
		FILE_READ_DATA or FILE_WRITE_DATA)

	IOCTL_CHANGEPTE	 equ CTL_CODE(FILE_DEVICE_UNKNOWN, CHANGEPTE, METHOD_BUFFERED, \
		FILE_READ_DATA or FILE_WRITE_DATA)

.code		

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	DriverEntry	proc, pDriverObject: dword, pRegPath: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	local	pDeviceObject:	dword	;pointer to dev obj created in IOCreateDevice
	
	;=== create named device object === 
	invoke IoCreateDevice, pDriverObject, 0, addr usDevice, FILE_DEVICE_UNKNOWN, \
		0, FALSE, addr pDeviceObject


.if eax == STATUS_SUCCESS

	;=== create a symbollic link so user processes can access device ===
	invoke IoCreateSymbolicLink, addr usSymbollic, addr usDevice

	.if eax == STATUS_SUCCESS

		mov eax, pDriverObject

            assume eax:PTR DRIVER_OBJECT
		mov [eax].MajorFunction+(IRP_MJ_CREATE*4), offset DispatchCreate
		mov [eax].MajorFunction+(IRP_MJ_CLOSE*4), offset DispatchClose 
		mov [eax].MajorFunction+(IRP_MJ_DEVICE_CONTROL*4), offset DispatchControl        
		mov [eax].DriverUnload, offset DriverUnload                                       
            assume eax:nothing

		mov eax, STATUS_SUCCESS
	.else

		;=== On failure delete device object ===
		invoke IoDeleteDevice, pDeviceObject
		mov eax, STATUS_DEVICE_CONFIGURATION_ERROR
	.endif

.else

	mov eax, STATUS_DEVICE_CONFIGURATION_ERROR

.endif


	
	ret

	

DriverEntry endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	DriverUnload	proc, pDriverObject: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	;=== delete symbollic link ===
	invoke IoDeleteSymbolicLink, addr usSymbollic

	;=== delete driver object ===
	mov eax, pDriverObject
     assume eax: PTR DRIVER_OBJECT
        invoke IoDeleteDevice, [eax].DeviceObject
     assume eax: nothing
	ret

DriverUnload endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	DispatchCreate	proc, pDeviceObject: dword, pIRP: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	;=== called when handle is created to our virtual device ===

	mov eax, pIRP
     assume eax: PTR _IRP
	mov [eax].IoStatus.Status, STATUS_SUCCESS
	mov [eax].IoStatus.Information, 0
     assume eax:nothing
	;=== let IO manager know everything is Okay ===
	invoke IoCompleteRequest, pIRP, IO_NO_INCREMENT

	mov eax, STATUS_SUCCESS

	ret

DispatchCreate endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	DispatchClose	proc, pDeviceObject: dword, pIRP: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	;=== called when handle is closed to our virtual device ===

	mov eax, pIRP
     assume eax: PTR _IRP
	mov [eax].IoStatus.Status, STATUS_SUCCESS
	mov [eax].IoStatus.Information, 0
     assume eax:nothing
	;=== let IO manager know everything is Okay ===
	invoke IoCompleteRequest, pIRP, IO_NO_INCREMENT

	mov eax, STATUS_SUCCESS

	ret

DispatchClose endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	DispatchControl proc, pDeviceObject: dword, pIRP: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	local	Status:		dword
	local	BytesReturned:	dword
	
	assume esi: PTR _IRP
	assume eax: PTR IO_STACK_LOCATION

	mov esi, pIRP
	mov eax, [esi].Tail.Overlay.CurrentStackLocation
	

	;================================================
	;======= Is it our GetPhys control code? ========
	;================================================

.if [eax].Parameters.DeviceIoControl.IoControlCode == IOCTL_GETPHYS


	;==== IN/OUT Buffer Usage ====
	;IN = Virtual
	;OUT = Physical


	;==========================================
	;== Are in and out buffer sizes 4 bytes? ==
	;==========================================

	.if ([eax].Parameters.DeviceIoControl.InputBufferLength == 4) && \
		([eax].Parameters.DeviceIoControl.OutputBufferLength == 4)
	  
		mov edi, [esi].AssociatedIrp.SystemBuffer	;pointer to in/out buffer
		mov ecx, [edi]
		pushad

		invoke MmGetPhysicalAddress, ecx

		mov BytesReturned, eax
		popad
		mov eax, BytesReturned


		mov [edi], eax
		mov Status, STATUS_SUCCESS
		mov BytesReturned, 4
	

	.else

		;=======================
		;== Buffers too small ==
		;=======================

		mov Status, STATUS_BUFFER_TOO_SMALL
		mov BytesReturned, 0
	.endif




	;==================================================
	;======= Is it our ChangePTE control code? ========
	;==================================================

.elseif [eax].Parameters.DeviceIoControl.IoControlCode == IOCTL_CHANGEPTE


	;=== IN/OUT Buffer usage ====
	;IN = Virtual (dword), Physical (dword)
	;OUT = Virtual (dword), old Physical (dword)


	;=============================================
	;== Are in and out buffers correctly sized? ==
	;=============================================

	.if ([eax].Parameters.DeviceIoControl.InputBufferLength == 8) && \
		([eax].Parameters.DeviceIoControl.OutputBufferLength == 8)
	  
		mov edi, [esi].AssociatedIrp.SystemBuffer	;pointer to in/out buffer
		
		pushad
			
		;==== In Arg's are Virtual and Physical in that order ====
		invoke ChangePTE, [edi], [edi+4]

		mov BytesReturned, eax
		popad
		mov eax, BytesReturned


		mov [edi+4], eax
		mov Status, STATUS_SUCCESS
		mov BytesReturned, 8		;return virtual and old physical address
	

	.else

		;=======================
		;== Buffers too small ==
		;=======================

		mov Status, STATUS_BUFFER_TOO_SMALL
		mov BytesReturned, 0
	.endif



.else

	;==========================
	;== Not our control code ==
	;==========================

	mov Status, STATUS_INVALID_DEVICE_REQUEST

.endif

	;==== Return Status =====
	mov eax, Status
	push eax
	mov [esi].IoStatus.Status, eax
	mov eax, BytesReturned
	mov [esi].IoStatus.Information, eax

	assume eax: nothing
	assume esi: nothing

	;=== let IO manager know  ===
	invoke IoCompleteRequest, pIRP, IO_NO_INCREMENT

	pop eax
	ret


DispatchControl endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	ChangePTE proc, Virtual: dword, Physical: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


	;== Virtual is address to modify in Page Table ==
	;== Physical is new physical address	       ==
	
	mov eax, Virtual

	shr eax, 22		;get PD index field
	shl eax, 2		;* (sizeof dword)
	
	add eax, 0c0300000h	;add base address of PD
				;eax points to PDE
	mov ecx, eax		;ecx points to PDE
	mov eax, [eax]		;eax = PDE for virtual address
	

.if (eax & 00000001b)		;check entry (P)resent flag

	.if (eax & 10000000b)	;check (P)age (S)ize flag
		
		;====Large Page (4mb)====
		push eax		;save PDE for return value
		and eax, 111111111111b	;preserve flags
		mov edx, Physical
		;round Physical to lower 4k page boundary
		and edx, 11111111111111111111000000000000b
		or edx, eax	;combine Physical with PDE flags
		mov [ecx], edx	;save new PDE
		
		pop eax
		and eax, 11111111111111111111000000000000b	;mask flags
	
	.else
		;====Small Page (4k)====
		mov eax, Virtual
		
		;Page Tables use VA's 0xC0000000-0xC03FFFFF.
		;Page Directory is at VA's 0xC0300000-0xC0300FFF.
		;The Physical address in the PDE is useless to us
		;since we can only use virtual addresses to access memory.
		;We have to combine the PD and the PT indexes to get the offset 
		;of the needed PT. The PDE for the 4mb space starting at 0xC0000000
		;points to the PD itself. The PD is used as a PT for these
		;addresses. This is how the PT's are successfully mapped at 
		;Virtual Address 0xC0000000. 
		
		shr eax, 12		;get combined index value	
		shl eax, 2		;* (sizeof dword)
		
		add eax, 0c0000000h		;add base address of PT's
		mov ecx, [eax]
			
			.if (ecx & 00000001b)	;check (P)resent flag
				
				push ecx		;save for return value
				and ecx, 111111111111b	;get PTE flags
				mov edx, Physical
				;round Physical to lower 4k page boundary
				and edx, 11111111111111111111000000000000b
				or edx, ecx	;combine Physical with PTE flags
				mov [eax], edx	;save new PTE
				pop eax
				;return original PTE physical address field
				and eax, 11111111111111111111000000000000b
		
			.else			;PTE not present
				
				xor eax, eax	;return 0

			.endif


	.endif

.else

		;Directory entry not present - swapped to disk, reserved or commited
		;with Virtualmem functions and not yet accessed, marked copy-on-write,
		;guard page, etc

		xor eax, eax

.endif

    ;==== Flush TLB ==== 
	mov edx, cr3
	mov cr3, edx

	ret

ChangePTE endp

;~~~~~~~~~~~~~~~
end DriverEntry
;~~~~~~~~~~~~~~~	