.586
.model flat, stdcall

include windows.inc
include macros.def
include kernel32.inc
include user32.inc

includelib kernel32.lib
includelib user32.lib
ShowMessage proto

comment	*

	Demonstrates use of MemPhys driver. 
	Alters process data by aliasing the ProcessHeap to the same physical
	memory as the data section and writing to it.

	*

;=========================================
;======== Includes From NtDDK ===========
;=========================================

CTL_CODE MACRO DeviceType:=<0>, Function:=<0>, Method:=<0>, Access:=<0>
	EXITM %(((DeviceType) SHL 16) OR ((Access) SHL 14) OR ((Function) SHL 2) OR (Method))
ENDM

FILE_DEVICE_UNKNOWN             equ 22h
METHOD_BUFFERED                 equ 0
FILE_READ_DATA				equ 0001
FILE_WRITE_DATA				equ 0002

;==== User Control Values ====

	GETPHYS = 800h
	CHANGEPTE = 801h


;==== Control Code Definitions =====

	IOCTL_GETPHYS	equ CTL_CODE(FILE_DEVICE_UNKNOWN, GETPHYS, METHOD_BUFFERED, \
		FILE_READ_DATA or FILE_WRITE_DATA)

	IOCTL_CHANGEPTE	 equ CTL_CODE(FILE_DEVICE_UNKNOWN, CHANGEPTE, METHOD_BUFFERED, \
		FILE_READ_DATA or FILE_WRITE_DATA)

VtoP struct
	Virtual		dd	?
	Physical	dd	?
VtoP ends

.data

	OriginalText	db	"This is before", NULL
			db	256	dup	(0)
	NewText		db	"And this is after", NULL

	cbNewText = $ - NewText

	NewPTE		VtoP	<>
	OldPTE		VtoP	<>
.code

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 	UseMemPhys	proc
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	local hFile:		dword	
	local BytesRet:		dword
	local hHeap:		dword

	;---- Open handle to MemPhys ----
	invoke CreateFile, SADD ("\\.\MemPhys"), GENERIC_READ or \
	   GENERIC_WRITE, NULL, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL
	

	.if eax == INVALID_HANDLE_VALUE

		invoke MessageBox, NULL, SADD ("MemPhys driver not installed."), \
		SADD ("Error"), MB_ICONSTOP
		jmp Exit
	.endif

	mov hFile, eax

	;---- Get physical address of process data section ----
	mov NewPTE.Virtual, offset OriginalText		
	invoke DeviceIoControl, hFile, IOCTL_GETPHYS, addr NewPTE.Virtual, 4, \
	 addr NewPTE.Physical, 4, addr BytesRet, NULL	


	;---- Handle to ProcessHeap is a pointer to committed page ----
	;---- within process address space			   ----
	invoke GetProcessHeap
	mov hHeap, eax		
	mov NewPTE.Virtual, eax		


	;---- Display text before change ----
	invoke ShowMessage


	;---- Change Page Table Entry ----
	invoke DeviceIoControl, hFile, IOCTL_CHANGEPTE, addr NewPTE, 8, addr OldPTE, \
		8, addr BytesRet, NULL


	;---- Copy text from process to Heap (mapped view of our data section) ----
	mov ecx, cbNewText
	mov esi, offset NewText		;text to copy
	mov edi, hHeap		;pointer to process heap, mapped to same
				;physical address as our data section.
				;When we write to the Heap, we are really
				;writing to the Physical memory of our Data section.
	cld 
	rep movsb
		
	;---- Restore Page Table Entry ----
	invoke DeviceIoControl, hFile, IOCTL_CHANGEPTE, addr OldPTE, 8, addr NewPTE, \
		8, addr BytesRet, NULL


	;---- Display changed text ----
	invoke ShowMessage

	
	invoke CloseHandle, hFile	
Exit:
	invoke ExitProcess, NULL



UseMemPhys endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;Displaying the Data sections text is isolated in a seperate procedure to
;demonstrate the fact that the data section has in fact been changed by 
;writing to the address of the ProcessHeap. This proc is called twice and each
;time it displays a different message.
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ShowMessage proc
	invoke MessageBox, NULL, addr OriginalText, SADD ("MemPhys Demo"), MB_OK
	ret
ShowMessage endp


;~~~~~~~~~~~
 end UseMemPhys
;~~~~~~~~~~~
